use crate::{
    decrypt_file_buffered, encrypt_file_buffered,
    errors::{CvfsErrors, CvfsResult},
    metadata::{FileInfo, FileTag, Metadata},
    open_rwlocked,
};
use multi_key_manager::{file::File, LiveKeyMaterial, MutableNonce, OpeningNonce};
use ring::aead::{OpeningKey, SealingKey};
use std::{
    io::{Read, Seek, Write},
    path::Path,
    sync::{Arc, Mutex, MutexGuard, Weak},
    thread::panicking,
};

/// an open active file
#[derive(Debug)]
pub struct ActiveFile {
    /// Clonable needed for metadata
    pub(crate) id: FileTag,
    /// The actual data buffer
    data: Vec<u8>,
    /// seekable
    offset: usize,
    /// We don't have the key inside this struct, but the `Metadata`
    modified_flag: bool,
    /// Locked file
    real_file: File,
}
impl ActiveFile {
    pub fn open_rwlocked(
        real_path: &Path,
        id: FileTag,
        file_key: &mut OpeningKey<OpeningNonce>,
    ) -> CvfsResult<Self> {
        let mut real_file = open_rwlocked(real_path)?;

        let data = if real_file.empty()? {
            Vec::new()
        } else {
            decrypt_file_buffered(&mut real_file, file_key)
                .map_err(|_generic_ring_error| CvfsErrors::LeafDecryptionError)?
        };

        Ok(Self {
            id,
            data,
            offset: 0,
            modified_flag: false,
            real_file,
        })
    }

    /// Same as `Write::flush` with more detailed error message
    pub fn try_flush(&mut self, key: &mut SealingKey<MutableNonce>) -> CvfsResult<()> {
        if self.data.is_empty() {
            // Don't save empty buffers as it confuses the decryption
            return Ok(());
        }
        encrypt_file_buffered(&mut self.real_file, &self.data, key)?;
        self.modified_flag = false;
        Ok(())
    }
}
impl Drop for ActiveFile {
    fn drop(&mut self) {
        if self.modified_flag {
            let error = "aead_vfs::ActiveFile was not flushed by Metadata";
            if panicking() {
                eprintln!("{error}");
            } else {
                panic!("{error}");
            }
        }
    }
}
impl Write for ActiveFile {
    /// All writes are to the buffer and not commited to disk until `.flush()?` is called or the file closes
    fn write(&mut self, buffer: &[u8]) -> Result<usize, std::io::Error> {
        let bytes_written = buffer.len();
        let new_max_len = self.offset + buffer.len();
        if self.data.len() < new_max_len {
            self.data.resize_with(new_max_len, Default::default);
        }
        let data_len = self.data.len();
        let target_range = self.offset..new_max_len;
        println!(" < target_range: {target_range:?}");
        let target = &mut self.data[target_range];
        println!(
            " < offset: {}, new: {new_max_len}, self.data: {data_len} target: {}, buffer: {}",
            self.offset,
            target.len(),
            buffer.len()
        );
        println!(" < target: {:?}", target);
        stdout_flush()?;

        // TODO try the more efficient method, while avoiding SIG-6
        //target.copy_from_slice(buffer);
        //target.clone_from_slice(buffer);
        // safes because it checks both ranges
        for (t, s) in target.iter_mut().zip(buffer.iter()) {
            *t = *s;
        }

        self.offset = new_max_len;
        self.modified_flag = true;

        Ok(bytes_written)
    }
    /// Will always panic!
    ///
    /// can not flush directly, must be done through the `SecureFileHandler`
    fn flush(&mut self) -> Result<(), std::io::Error> {
        panic!("can not flush directly, must be done through the SecureFileHandler")
    }
}
impl Seek for ActiveFile {
    fn seek(&mut self, target: std::io::SeekFrom) -> Result<u64, std::io::Error> {
        use std::io::SeekFrom::*;
        self.offset = match target {
            Start(target) => target.try_into().expect("unable to convert seeking value"),
            Current(target) => {
                let new = self.offset as i64 + target;
                if new < 0 {
                    panic!("seeked before start");
                }
                new as usize
            }
            End(target) => {
                let new = self.data.len() as i64 + target;
                if new < 0 {
                    panic!("seeked before start");
                }
                new as usize
            }
        };
        Ok(self.offset as u64)
    }
}

/// These Handlers should reduce the time the main Mutex is held so concurrent access is possible
pub struct SecureFileHandler {
    pub(crate) metadata: Weak<Mutex<Metadata>>,
    pub(crate) id: FileTag,
}
impl SecureFileHandler {
    fn try_write(&mut self, buffer: &[u8]) -> CvfsResult<usize> {
        let arc = self
            .metadata
            .upgrade()
            .expect("Metadata was dropped before SecureFileHandler");
        let mut metadata = arc.lock().expect("Metadata Mutex poisened");
        let info = metadata
            .find_info(&self.id)
            .expect("Cvfs dropped before SecureFileHandler::try_write");

        todo!()
    }
    fn try_flush(&mut self) -> CvfsResult<()> {
        let arc = self
            .metadata
            .upgrade()
            .expect("Metadata was dropped before SecureFileHandler");
        let mut metadata = arc.lock().expect("Metadata Mutex poisened");
        let mut info = metadata
            .find_info(&self.id)
            .expect("Cvfs dropped before SecureFileHandler::try_flush");

        todo!()
    }
}
impl Write for SecureFileHandler {
    fn write(&mut self, buffer: &[u8]) -> Result<usize, std::io::Error> {
        self.try_write(buffer)
            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, format!("{e:?}")))
    }
    fn flush(&mut self) -> Result<(), std::io::Error> {
        self.try_flush()
            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, format!("{e:?}")))
    }
}

fn stdout_flush() -> std::io::Result<()> {
    std::io::stdout().flush()
}

#[cfg(test)]
mod tests {
    use super::*;
    use ring::aead::{BoundKey, UnboundKey};
    use std::{cell::RefCell, path::PathBuf, rc::Rc};

    /// ext must not start with a `/`
    fn path<P: AsRef<Path>>(ext: P) -> PathBuf {
        let mut p = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
        // we are in a cargo workspace
        p.pop();
        p.push("target");
        p.push("aead-vfs_tests-secure-file");
        std::fs::create_dir_all(&p).expect("unable to create testing directory");
        p.push(ext);
        p
    }

    fn test_key() -> UnboundKey {
        let key_material = [42u8; 32];
        let algorithm = &ring::aead::CHACHA20_POLY1305;

        UnboundKey::new(algorithm, &key_material).expect("unable to construct UnboundKey")
    }
    fn test_opening_key() -> OpeningKey<OpeningNonce> {
        let nonce = OpeningNonce::new(2);
        OpeningKey::new(test_key(), nonce)
    }
    fn test_sealing_key() -> SealingKey<MutableNonce> {
        let nonce = MutableNonce::new(Rc::new(RefCell::new(1u128)));
        SealingKey::new(test_key(), nonce)
    }

    #[test]
    fn simple_write() {
        let path = path("simple_write.bin");
        // delete if it already exists
        let _ = std::fs::remove_file(&path);
        let mut key = test_opening_key();
        let mut a = ActiveFile::open_rwlocked(&path, FileTag(1), &mut key)
            .expect("unable to open test file");
        assert_eq!(0, a.offset);

        a.write(b"abc").unwrap();
        assert_eq!(true, a.modified_flag);
        println!("a.data: {:?}", a.data);
        assert_eq!(3, a.data.len());
        assert_eq!(3, a.offset);

        a.write(b"def").unwrap();
        println!("a.data: {:?}", a.data);
        assert_eq!(6, a.data.len());
        assert_eq!(6, a.offset);

        let mut key = test_sealing_key();
        a.try_flush(&mut key).expect("unable to flush");
    }

    #[test]
    fn write_and_seek() {
        let mut key = test_opening_key();
        let mut a = ActiveFile::open_rwlocked(&path("write_and_seek.bin"), FileTag(1), &mut key)
            .expect("unable to open test file");

        a.write(b"abcdefghi").unwrap();
        println!("a.data: {:?}", a.data);
        assert_eq!(9, a.data.len());
        assert_eq!(9, a.offset);

        a.seek(std::io::SeekFrom::Current(-6));
        a.write(b"XYZ").unwrap();
        assert_eq!(true, a.modified_flag);
        println!("a.data: {:?}", a.data);
        assert_eq!(9, a.data.len());
        assert_eq!(6, a.offset);

        let mut key = test_sealing_key();
        a.try_flush(&mut key).unwrap();
        assert_eq!(false, a.modified_flag);
    }
}
